## Copyright (C) 2012 by Kevin L. Mitchell <klmitch@mit.edu>
##
## This program is free software: you can redistribute it and/or
## modify it under the terms of the GNU General Public License as
## published by the Free Software Foundation, either version 3 of the
## License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program.  If not, see
## <http://www.gnu.org/licenses/>.

import inspect
import unittest

import mock

from tendril import framers


class TestFrameState(unittest.TestCase):
    def test_init(self):
        state = framers.FrameState()

        self.assertEqual(state.recv_buf, '')
        self.assertEqual(state.send_buf, [])
        self.assertEqual(state._other, {})
        self.assertEqual(state._framer_id, None)

    def test_getattr(self):
        state = framers.FrameState()
        state._other.update(a=1, _b=2)

        self.assertEqual(state.recv_buf, '')
        self.assertEqual(state.send_buf, [])
        self.assertEqual(state.a, 1)
        with self.assertRaises(AttributeError):
            dummy = state.b
        with self.assertRaises(AttributeError):
            dummy = state._b

    def test_setattr(self):
        state = framers.FrameState()

        state.recv_buf = 'foobar'
        state.send_buf = [1, 2, 3]
        state._framer_id = 'foo'
        state.a = 1

        self.assertEqual(state.recv_buf, 'foobar')
        self.assertEqual(state.send_buf, [1, 2, 3])
        self.assertEqual(state._framer_id, 'foo')
        self.assertEqual(state._other, dict(a=1))

    def test_delattr(self):
        state = framers.FrameState()
        state._other.update(a=1, _b=2)

        with self.assertRaises(AttributeError):
            del state.recv_buf
        with self.assertRaises(AttributeError):
            del state.send_buf
        with self.assertRaises(AttributeError):
            del state._b
        del state.a
        self.assertEqual(state._other, dict(_b=2))
        with self.assertRaises(AttributeError):
            del state.a

    def test_reset(self):
        state = framers.FrameState()
        state.recv_buf = 'foobar'
        state.send_buf = [1, 2, 3]
        state._other.update(a=1, b=2, c=3)

        mock_framer = mock.Mock()
        state._reset(mock_framer)

        self.assertEqual(state.recv_buf, 'foobar')
        self.assertEqual(state.send_buf, [1, 2, 3])
        self.assertEqual(state._other, {})
        self.assertEqual(state._framer_id, id(mock_framer))
        mock_framer.init_state.assert_called_once_with(state)

        state.recv_buf += 'baz'
        state.send_buf.append(4)
        state.a = 3
        state.b = 2
        state.c = 1

        mock_framer.reset_mock()
        state._reset(mock_framer)

        self.assertEqual(state.recv_buf, 'foobarbaz')
        self.assertEqual(state.send_buf, [1, 2, 3, 4])
        self.assertEqual(state._other, dict(a=3, b=2, c=1))
        self.assertEqual(state._framer_id, id(mock_framer))
        self.assertFalse(mock_framer.init_state.called)


class TestFramer(unittest.TestCase):
    framer_class = framers.Framer
    clear_state = {}

    def test_has_all_methods(self):
        self.assertTrue(hasattr(self.framer_class, 'init_state'))
        self.assertTrue(callable(self.framer_class.init_state))
        self.assertTrue(hasattr(self.framer_class, 'frameify'))
        self.assertTrue(callable(self.framer_class.frameify))
        self.assertTrue(hasattr(self.framer_class, 'streamify'))
        self.assertTrue(callable(self.framer_class.streamify))

    def test_framify_is_generator(self):
        self.assertTrue(inspect.isgeneratorfunction(
            self.framer_class.frameify))

    def check_init_state(self, *args, **kwargs):
        # Instantiate the framer...
        f = self.framer_class(*args, **kwargs)

        # Get a frame state
        state = framers.FrameState()
        state._reset(f)

        # Confirm the state is correct
        self.assertEqual(state.recv_buf, '')
        self.assertEqual(state.send_buf, [])
        self.assertEqual(state._other, self.clear_state)

    def check_interrupt(self, stream, frame, rest, *args, **kwargs):
        # Instantiate the framer...
        f = self.framer_class(*args, **kwargs)

        # Also need a frame state
        state = framers.FrameState()
        state._reset(f)

        # Set up the generator
        gen = f.frameify(state, stream)

        # Get one frame
        fr1 = gen.next()

        self.assertEqual(fr1, frame)

        # Interrupt the framer
        with self.assertRaises(StopIteration):
            gen.throw(framers.FrameSwitch())

        self.assertEqual(state.recv_buf, rest)

    def check_composition(self, frames, *args, **kwargs):
        # Instantiate the framer...
        f = self.framer_class(*args, **kwargs)

        # Also need a frame state
        state = framers.FrameState()
        state._reset(f)

        # Convert the incoming frames into a stream
        stream = ''
        for frame in frames:
            stream += f.streamify(state, frame)

        # Now go the other way
        gen = f.frameify(state, stream)
        self.assertTrue(inspect.isgenerator(gen))
        results = list(gen)

        # Confirm that the state has been reset
        self.assertEqual(state.recv_buf, '')
        self.assertEqual(state.send_buf, [])
        self.assertEqual(state._other, self.clear_state)
        self.assertEqual(frames, results)


class TestIdentityFramer(TestFramer):
    framer_class = framers.IdentityFramer

    def test_init_state(self):
        self.check_init_state()

    def test_frameify(self):
        f = self.framer_class()
        s = framers.FrameState()
        s._reset(f)

        result = f.frameify(s, 'this is a test')

        self.assertEqual(list(result), ['this is a test'])
        self.assertEqual(s.recv_buf, '')

    def test_frameify_buffered(self):
        f = self.framer_class()
        s = framers.FrameState()
        s._reset(f)
        s.recv_buf = 'this is '

        result = f.frameify(s, 'a test')

        self.assertEqual(list(result), ['this is a test'])
        self.assertEqual(s.recv_buf, '')

    def test_frameify_interrupt(self):
        self.check_interrupt('this is a test', 'this is a test', '')

    def test_streamify(self):
        f = self.framer_class()

        result = f.streamify(None, 'this is a test')

        self.assertEqual(result, 'this is a test')

    def test_composition(self):
        self.check_composition(['this is a test'])


class TestChunkFramer(TestFramer):
    framer_class = framers.ChunkFramer
    clear_state = dict(chunk_remaining=20)

    def test_init(self):
        f = self.framer_class(255)

        self.assertEqual(f.chunk_len, 255)

    def test_init_state(self):
        self.check_init_state(20)

    def test_frameify(self):
        f = self.framer_class(10)
        s = framers.FrameState()
        s._reset(f)

        result = f.frameify(s, '1234567890')

        self.assertEqual(list(result), ['1234567890'])
        self.assertEqual(s.recv_buf, '')
        self.assertEqual(s._other, dict(chunk_remaining=0))

    def test_frameify_complete(self):
        f = self.framer_class(5)
        s = framers.FrameState()
        s._reset(f)

        result = f.frameify(s, '1234567890')

        self.assertEqual(list(result), ['12345'])
        self.assertEqual(s.recv_buf, '67890')
        self.assertEqual(s._other, dict(chunk_remaining=0))

    def test_frameify_buffered(self):
        f = self.framer_class(20)
        s = framers.FrameState()
        s._reset(f)
        s.recv_buf = '12345'

        result = f.frameify(s, '67890')

        self.assertEqual(list(result), ['1234567890'])
        self.assertEqual(s.recv_buf, '')
        self.assertEqual(s._other, dict(chunk_remaining=10))

    def test_frameify_buffering(self):
        f = self.framer_class(0)
        s = framers.FrameState()
        s._reset(f)
        s.recv_buf = '12345'

        result = f.frameify(s, '67890')

        self.assertEqual(list(result), [])
        self.assertEqual(s.recv_buf, '1234567890')
        self.assertEqual(s._other, dict(chunk_remaining=0))

    @mock.patch.dict(clear_state, chunk_remaining=0)
    def test_frameify_interrupt(self):
        self.check_interrupt('1234567890', '12345', '67890', 5)

    @mock.patch.dict(clear_state, chunk_remaining=0)
    def test_composition(self):
        self.check_composition(['1234567890'], 10)


class TestLineFramer(TestFramer):
    framer_class = framers.LineFramer

    def test_init(self):
        f1 = self.framer_class()
        f2 = self.framer_class(False)

        self.assertEqual(f1.carriage_return, True)
        self.assertEqual(f2.carriage_return, False)
        self.assertEqual(f1.line_end, '\r\n')
        self.assertEqual(f2.line_end, '\n')

    def test_frameify(self):
        f = self.framer_class()
        s = framers.FrameState()
        s._reset(f)

        result = f.frameify(s, 'frame1\nframe2\r\n')

        self.assertEqual(list(result), ['frame1', 'frame2'])
        self.assertEqual(s.recv_buf, '')

    def test_frameify_nocr(self):
        f = self.framer_class(False)
        s = framers.FrameState()
        s._reset(f)

        result = f.frameify(s, 'frame1\nframe2\r\n')

        self.assertEqual(list(result), ['frame1', 'frame2\r'])
        self.assertEqual(s.recv_buf, '')

    def test_frameify_buffered(self):
        f = self.framer_class()
        s = framers.FrameState()
        s._reset(f)
        s.recv_buf = 'this is '

        result = f.frameify(s, 'frame1\nframe2')

        self.assertEqual(list(result), ['this is frame1'])
        self.assertEqual(s.recv_buf, 'frame2')

    def test_frameify_interrupt(self):
        self.check_interrupt('frame1\nframe2\n', 'frame1', 'frame2\n')

    def test_streamify(self):
        f = self.framer_class()

        result = f.streamify(None, 'this is a test')

        self.assertEqual(result, 'this is a test\r\n')

    def test_streamify_nocr(self):
        f = self.framer_class(False)

        result = f.streamify(None, 'this is a test')

        self.assertEqual(result, 'this is a test\n')

    def test_composition(self):
        self.check_composition(['frame1', 'frame2', 'frame3', 'frame4'])

    def test_composition_nocr(self):
        self.check_composition(['frame1', 'frame2', 'frame3', 'frame4'])


class TestStructFramer(TestFramer):
    framer_class = framers.StructFramer
    clear_state = dict(frame_len=None)

    def make_frame(self, text, length=None):
        return chr(len(text) if length is None else length) + text

    def test_init(self):
        self.assertRaises(ValueError, self.framer_class, 'z')
        self.assertRaises(ValueError, self.framer_class, 'ii')
        self.assertRaises(ValueError, self.framer_class, '!')

        f = self.framer_class('!B')

        self.assertEqual(f.fmt.format, '!B')

    def test_init_state(self):
        self.check_init_state('!B')

    def test_frameify(self):
        f = self.framer_class('!B')
        s = framers.FrameState()
        s._reset(f)

        result = f.frameify(s, ''.join([
            self.make_frame('frame1'),
            self.make_frame('frame2'),
        ]))

        self.assertEqual(list(result), ['frame1', 'frame2'])
        self.assertEqual(s.recv_buf, '')
        self.assertEqual(s._other, self.clear_state)

    def test_frameify_buffered(self):
        f = self.framer_class('!B')
        s = framers.FrameState()
        s._reset(f)
        s.recv_buf = self.make_frame('sp', 5)

        result = f.frameify(s, ''.join(['am1', self.make_frame('sp', 5)]))

        self.assertEqual(list(result), ['spam1'])
        self.assertEqual(s.recv_buf, 'sp')
        self.assertEqual(s._other, dict(frame_len=5))

    def test_frameify_fractured(self):
        f = self.framer_class('!B')
        s = framers.FrameState()
        s._reset(f)
        frame = self.make_frame('spam', 4)
        s.frame_len = 4

        result = f.frameify(s, frame[1:])

        self.assertEqual(list(result), ['spam'])
        self.assertEqual(s.recv_buf, '')
        self.assertEqual(s._other, self.clear_state)

    def test_frameify_buffered_len(self):
        f = self.framer_class('<i')
        s = framers.FrameState()
        s._reset(f)

        result = f.frameify(s, self.make_frame('\0', 4))

        self.assertEqual(list(result), [])
        self.assertEqual(s.recv_buf, self.make_frame('\0', 4))
        self.assertEqual(s._other, self.clear_state)

    def test_frameify_interrupt(self):
        self.check_interrupt(''.join([
            self.make_frame('frame1'),
            self.make_frame('frame2'),
        ]), 'frame1', self.make_frame('frame2'), '!B')

    def test_streamify_byte(self):
        f = self.framer_class('!B')

        result = f.streamify(None, 'this is a test')

        self.assertEqual(result, self.make_frame('this is a test'))

    def test_streamify_int(self):
        f = self.framer_class('>i')

        result = f.streamify(None, 'this is a test')

        self.assertEqual(result, '\0\0\0' + self.make_frame('this is a test'))

    def test_composition(self):
        self.check_composition(['frame1', 'frame2', 'frame3', 'frame4'], '!B')


class TestStuffingFramer(TestFramer):
    framer_class = framers.StuffingFramer
    clear_state = dict(frame_start=False)

    def test_init(self):
        self.assertRaises(ValueError, self.framer_class, begin='aa')
        self.assertRaises(ValueError, self.framer_class, end='aa')
        self.assertRaises(ValueError, self.framer_class, nop='aa')
        self.assertRaises(ValueError, self.framer_class, begin='a', nop='a')
        self.assertRaises(ValueError, self.framer_class, end='a', nop='a')
        self.assertRaises(ValueError, self.framer_class, begin='a', end='a')

        f = self.framer_class('aaa', begin='a', end='b', nop='c')

        self.assertEqual(f.prefix, 'aaa')
        self.assertEqual(f.begin, 'a')
        self.assertEqual(f.end, 'b')
        self.assertEqual(f.nop, 'c')

    def test_init_state(self):
        self.check_init_state()

    def test_frameify(self):
        f = self.framer_class(prefix='zzz', begin='z', end='w', nop='a')
        s = framers.FrameState()
        s._reset(f)

        result = f.frameify(s, 'zzzzthis is a testzzzwzzzzdid it work?zzzw')

        self.assertEqual(list(result), ['this is a test', 'did it work?'])
        self.assertEqual(s.recv_buf, '')
        self.assertEqual(s._other, self.clear_state)

    def test_frameify_started(self):
        f = self.framer_class(prefix='zzz', begin='z', end='w', nop='a')
        s = framers.FrameState()
        s._reset(f)
        s.frame_start = True

        result = f.frameify(s, 'zzzzthis is a testzzzw')

        self.assertEqual(list(result), ['zzzzthis is a test'])
        self.assertEqual(s.recv_buf, '')
        self.assertEqual(s._other, self.clear_state)

    def test_frameify_nop(self):
        f = self.framer_class(prefix='zzz', begin='z', end='w', nop='a')
        s = framers.FrameState()
        s._reset(f)

        result = f.frameify(s, 'zzzzthis is a zzzazzzzw')

        self.assertEqual(list(result), ['this is a zzzz'])
        self.assertEqual(s.recv_buf, '')
        self.assertEqual(s._other, self.clear_state)

    def test_frameify_buffered_nostart(self):
        f = self.framer_class(prefix='zzz', begin='z', end='w', nop='a')
        s = framers.FrameState()
        s._reset(f)
        s.recv_buf = 'zzz'

        result = f.frameify(s, 'zthis is a testzzzwzzz')

        self.assertEqual(list(result), ['this is a test'])
        self.assertEqual(s.recv_buf, 'zzz')
        self.assertEqual(s._other, self.clear_state)

    def test_frameify_buffered_noend(self):
        f = self.framer_class(prefix='zzz', begin='z', end='w', nop='a')
        s = framers.FrameState()
        s._reset(f)

        result = f.frameify(s, 'zzzzthis is a testzzzwzzzzdid it work?zzz')

        self.assertEqual(list(result), ['this is a test'])
        self.assertEqual(s.recv_buf, 'did it work?zzz')
        self.assertEqual(s._other, dict(frame_start=True))

    def test_frameify_interrupt(self):
        self.check_interrupt('zzzzthis is a testzzzwzzzzdid it work?zzzw',
                             'this is a test', 'zzzzdid it work?zzzw',
                             prefix='zzz', begin='z', end='w', nop='a')

    def test_streamify(self):
        f = self.framer_class(prefix='zzz', begin='z', end='w', nop='a')

        result = f.streamify(None, 'this is a zzzz so I wonder zzzzzzzzzzz '
                             'if it worked')

        self.assertEqual(result, 'zzzzthis is a zzzaz so I wonder '
                         'zzzazzzazzzazz if it workedzzzw')

    def test_composition(self):
        self.check_composition([
            'test one',
            'test zzzzzzzzzzzw two',
            'zzzz test three zzzw',
        ])


class TestCOBSFramer(TestFramer):
    framer_class = framers.COBSFramer

    def test_init(self):
        f1 = self.framer_class()
        f2 = self.framer_class(True)

        self.assertEqual(f1.zpe, False)
        self.assertEqual(f2.zpe, True)

    def test_frameify(self):
        f = self.framer_class()
        s = framers.FrameState()
        s._reset(f)

        result = f.frameify(s, '\x01\x01\x00'
                            '\x03\x11\x22\x02\x33\x00'
                            '\x02\x11\x01\x01\x01\x00'
                            '\xff' + ''.join(chr(i) for i in range(1, 255)) +
                            '\x02\xff\x00')

        self.assertEqual(list(result), [
            '\x00',
            '\x11\x22\x00\x33',
            '\x11\x00\x00\x00',
            ''.join(chr(i) for i in range(1, 256)),
        ])
        self.assertEqual(s.recv_buf, '')
        self.assertEqual(s._other, self.clear_state)

    def test_frameify_buffered(self):
        f = self.framer_class()
        s = framers.FrameState()
        s._reset(f)
        s.recv_buf = '\xff' + ''.join(chr(i) for i in range(1, 255))

        result = f.frameify(s, '\x02\xff\x00\x01\x01')

        self.assertEqual(list(result), [
            ''.join(chr(i) for i in range(1, 256)),
        ])
        self.assertEqual(s.recv_buf, '\x01\x01')
        self.assertEqual(s._other, self.clear_state)

    def test_frameify_interrupt(self):
        self.check_interrupt('\x01\x01\x00\x03\x11\x22\x02\x33\x00',
                             '\x00', '\x03\x11\x22\x02\x33\x00')

    def test_frameify_zpe(self):
        f = self.framer_class(True)
        s = framers.FrameState()
        s._reset(f)

        result = f.frameify(s, '\xe1\x00'
                            '\x03\x11\x22\x02\x33\x00'
                            '\xe2\x11\xe1\x00'
                            '\xe0' + ''.join(chr(i) for i in range(1, 224)) +
                            '\x21' + ''.join(chr(i) for i in range(224, 256)) +
                            '\x00')

        self.assertEqual(list(result), [
            '\x00',
            '\x11\x22\x00\x33',
            '\x11\x00\x00\x00',
            ''.join(chr(i) for i in range(1, 256)),
        ])
        self.assertEqual(s.recv_buf, '')
        self.assertEqual(s._other, self.clear_state)

    def test_frameify_zpe_buffered(self):
        f = self.framer_class(True)
        s = framers.FrameState()
        s._reset(f)
        s.recv_buf = '\xe0' + ''.join(chr(i) for i in range(1, 224))

        result = f.frameify(s, '\x21' +
                            ''.join(chr(i) for i in range(224, 256)) +
                            '\x00\xe1')

        self.assertEqual(list(result), [
            ''.join(chr(i) for i in range(1, 256)),
        ])
        self.assertEqual(s.recv_buf, '\xe1')
        self.assertEqual(s._other, self.clear_state)

    def test_frameify_zpe_interrupt(self):
        self.check_interrupt('\xe1\x00\x03\x11\x22\x02\x33\x00',
                             '\x00', '\x03\x11\x22\x02\x33\x00', True)

    def test_streamify(self):
        f = self.framer_class()

        result = f.streamify(None, '\x00')

        self.assertEqual(result, '\x01\x01\x00')

        result = f.streamify(None, '\x11\x22\x00\x33')

        self.assertEqual(result, '\x03\x11\x22\x02\x33\x00')

        result = f.streamify(None, '\x11\x00\x00\x00')

        self.assertEqual(result, '\x02\x11\x01\x01\x01\x00')

        result = f.streamify(None, ''.join(chr(i) for i in range(1, 256)))

        self.assertEqual(result, '\xff' +
                         ''.join(chr(i) for i in range(1, 255)) +
                         '\x02\xff\x00')

    def test_streamify_zpe(self):
        f = self.framer_class(True)

        result = f.streamify(None, '\x00')

        self.assertEqual(result, '\xe1\x00')

        result = f.streamify(None, '\x11\x22\x00\x33')

        self.assertEqual(result, '\x03\x11\x22\x02\x33\x00')

        result = f.streamify(None, '\x11\x00\x00\x00')

        self.assertEqual(result, '\xe2\x11\xe1\x00')

        result = f.streamify(None, ''.join(chr(i) for i in range(1, 256)))

        self.assertEqual(result, '\xe0' +
                         ''.join(chr(i) for i in range(1, 224)) + '\x21' +
                         ''.join(chr(i) for i in range(224, 256)) + '\x00')

    def test_composition(self):
        self.check_composition([
            '\x00',
            '\x11\x22\x00\x33',
            '\x11\x00\x00\x00',
            ''.join(chr(i) for i in range(1, 256)),
        ])

    def test_composition_zpe(self):
        self.check_composition([
            '\x00',
            '\x11\x22\x00\x33',
            '\x11\x00\x00\x00',
            ''.join(chr(i) for i in range(1, 256)),
        ], True)



## Copyright (C) 2012 by Kevin L. Mitchell <klmitch@mit.edu>
##
## This program is free software: you can redistribute it and/or
## modify it under the terms of the GNU General Public License as
## published by the Free Software Foundation, either version 3 of the
## License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program.  If not, see
## <http://www.gnu.org/licenses/>.

import time
import unittest

import gevent
import mock

import tendril
from tendril import manager


class EchoApplicationClient(tendril.Application):
    def __init__(self, parent):
        super(EchoApplicationClient, self).__init__(parent)

        self.received_frames = []

    def recv_frame(self, frame):
        self.received_frames.append(frame)


def start_client(proto, server):
    manager = tendril.get_manager(proto)
    manager.start()

    tend = manager.connect(server, EchoApplicationClient)

    return manager, tend.application


class EchoApplicationServer(tendril.Application):
    def recv_frame(self, frame):
        self.send_frame(frame)


def start_server(proto):
    manager = tendril.get_manager(proto, ('127.0.0.1', 0))
    manager.start(EchoApplicationServer)
    return manager


class TestBasicFunction(unittest.TestCase):
    def setUp(self):
        self.p_managers = mock.patch.object(manager.TendrilManager,
                                            '_managers', {})
        self.p_tendrils = mock.patch.object(manager.TendrilManager,
                                            '_tendrils', {})
        self.p_running_managers = mock.patch.object(manager.TendrilManager,
                                                    '_running_managers', {})

        self.p_managers.start()
        self.p_tendrils.start()
        self.p_running_managers.start()

        self.serv = start_server(self.proto)
        self.cli, self.app = start_client(self.proto, self.serv.local_addr)

    def tearDown(self):
        self.serv.shutdown()
        self.cli.shutdown()

        self.p_running_managers.stop()
        self.p_tendrils.stop()
        self.p_managers.stop()

    def test_proto(self):
        # Send some messages
        self.app.send_frame('frame1')
        self.app.send_frame('frame2')
        self.app.send_frame('frame3')

        # Sleep to give the server a chance to handle them
        end = time.time() + 10
        while time.time() < end and len(self.app.received_frames) < 3:
            gevent.sleep(1)

        # Verify that the client got the messages back
        self.assertEqual(self.app.received_frames,
                         ['frame1', 'frame2', 'frame3'])



## Copyright (C) 2012 by Kevin L. Mitchell <klmitch@mit.edu>
##
## This program is free software: you can redistribute it and/or
## modify it under the terms of the GNU General Public License as
## published by the Free Software Foundation, either version 3 of the
## License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program.  If not, see
## <http://www.gnu.org/licenses/>.

from tests.function import base


class TestUDPFunction(base.TestBasicFunction):
    proto = 'udp'



## Copyright (C) 2012 by Kevin L. Mitchell <klmitch@mit.edu>
##
## This program is free software: you can redistribute it and/or
## modify it under the terms of the GNU General Public License as
## published by the Free Software Foundation, either version 3 of the
## License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program.  If not, see
## <http://www.gnu.org/licenses/>.

import gevent
from gevent import coros
from gevent import event
from gevent import socket

from tendril import application
from tendril import connection
from tendril import framers
from tendril import manager
from tendril import utils


class TCPTendril(connection.Tendril):
    """
    Manages state associated with a single TCP connection.  In
    addition to the attributes available on the Tendril class, this
    class includes the attribute ``recv_bufsize``, which allows tuning
    of the amount of data requested from the system for each call to
    the socket ``recv()`` method.  This class also provides the
    property ``sock``, allowing access to the underlying ``socket``
    object (or its wrapper).
    """

    default_framer = framers.LineFramer
    proto = 'tcp'
    recv_bufsize = 4096

    def __init__(self, manager, sock, remote_addr=None):
        """
        Initialize a TCPTendril.

        :param manager: The TCPTendrilManager responsible for the
                        Tendril.
        :param sock: The socket for the underlying connection.
        :param remote_addr: The address of the remote end of the
                            connection represented by the TCPTendril.
                            If not provided, will be derived by
                            calling the ``getpeername()`` method on
                            the socket.
        """

        super(TCPTendril, self).__init__(manager,
                                         sock.getsockname(),
                                         remote_addr or sock.getpeername())

        self._sock = sock

        # Send buffer and support
        self._sendbuf_event = event.Event()
        self._sendbuf = ''

        # Thread objects for the send and receive threads
        self._recv_thread = None
        self._send_thread = None

        # Locks for the send and receive threads
        self._recv_lock = None
        self._send_lock = None

    def _start(self):
        """
        Starts the underlying send and receive threads.
        """

        # Initialize the locks
        self._recv_lock = coros.Semaphore(0)
        self._send_lock = coros.Semaphore(0)

        # Boot the threads
        self._recv_thread = gevent.spawn(self._recv)
        self._send_thread = gevent.spawn(self._send)

        # Link the threads such that we get notified if one or the
        # other exits
        self._recv_thread.link(self._thread_error)
        self._send_thread.link(self._thread_error)

    def _recv(self):
        """
        Implementation of the receive thread.  Waits for data to
        arrive on the socket, then passes the data through the defined
        receive framer and sends it on to the application.
        """

        # Outer loop: receive some data
        while True:
            # Wait until we can go
            self._recv_lock.release()
            gevent.sleep()  # Yield to another thread
            self._recv_lock.acquire()

            recv_buf = self._sock.recv(self.recv_bufsize)

            # If it's empty, the peer closed the other end
            if not recv_buf:
                # Manually kill the send thread; do this manually
                # instead of calling close() because close() will kill
                # us, and since close() would be running in our thread
                # context, it would never get around to killing the
                # send thread
                if self._send_thread:
                    self._send_thread.kill()
                    self._send_thread = None

                # Manually close the socket
                self._sock.close()
                self._sock = None

                # Make sure the manager knows we're closed
                super(TCPTendril, self).close()

                # Notify the application
                self.closed()

                # As our last step, commit seppuku; this will keep
                # _thread_error() from notifying the application of an
                # erroneous exit from the receive thread
                raise gevent.GreenletExit()

            # Process the received data
            self._recv_frameify(recv_buf)

    def _send(self):
        """
        Implementation of the send thread.  Waits for data to be added
        to the send buffer, then passes as much of the buffered data
        to the socket ``send()`` method as possible.
        """

        # Outer loop: wait for data to send
        while True:
            # Release the send lock and wait for data, then reacquire
            # the send lock
            self._send_lock.release()
            self._sendbuf_event.wait()
            self._send_lock.acquire()

            # Inner loop: send as much data as we can
            while self._sendbuf:
                sent = self._sock.send(self._sendbuf)

                # Trim that much data off the send buffer, so we don't
                # accidentally re-send anything
                self._sendbuf = self._sendbuf[sent:]

            # OK, _sendbuf is empty; clear the event so we'll sleep
            self._sendbuf_event.clear()

    def _thread_error(self, thread):
        """
        Handles the case that the send or receive thread exit or throw
        an exception.
        """

        # Avoid double-killing the thread
        if thread == self._send_thread:
            self._send_thread = None
        if thread == self._recv_thread:
            self._recv_thread = None

        # Figure out why the thread exited
        if thread.successful():
            exception = socket.error('thread exited prematurely')
        elif isinstance(thread.exception, gevent.GreenletExit):
            # Thread was killed; don't do anything but close
            self.close()
            return
        else:
            exception = thread.exception

        # Close the connection...
        self.close()

        # Notify the application what happened
        self.closed(exception)

    def wrap(self, wrapper):
        """
        Allows the underlying socket to be wrapped, as by an SSL
        connection.

        :param wrapper: A callable taking, as its first argument, a
                        socket.socket object.  The callable must
                        return a valid proxy for the socket.socket
                        object, which will subsequently be used to
                        communicate on the connection.

        Note: Be extremely careful with calling this method after the
        TCP connection has been initiated.  The action of this method
        affects both sending and receiving streams simultaneously, and
        no attempt is made to deal with buffered data, other than
        ensuring that both the sending and receiving threads are at
        stopping points.
        """

        if self._recv_thread and self._send_thread:
            # Have to suspend the send/recv threads
            self._recv_lock.acquire()
            self._send_lock.acquire()

        # Wrap the socket
        self._sock = wrapper(self._sock)

        # OK, restart the send/recv threads
        if self._recv_thread and self._send_thread:
            # Release our locks
            self._send_lock.release()
            self._recv_lock.release()

    @property
    def sock(self):
        """
        Retrieve the underlying socket object.  Read-only.
        """

        return self._sock

    def send_frame(self, frame):
        """
        Sends a frame to the other end of the connection.
        """

        self._sendbuf += self._send_streamify(frame)
        self._sendbuf_event.set()

    def close(self):
        """
        Close the connection.  Kills the send and receive threads, as
        well as closing the underlying socket.
        """

        if self._recv_thread:
            self._recv_thread.kill()
            self._recv_thread = None

        if self._send_thread:
            self._send_thread.kill()
            self._send_thread = None

        if self._sock:
            self._sock.close()
            self._sock = None

        # Make sure to notify the manager we're closed
        super(TCPTendril, self).close()


class TCPTendrilManager(manager.TendrilManager):
    """
    Manages new connections through a particular endpoint.  Handles
    accepting new connections and creating new outgoing connections.
    This class includes the attribute ``backlog``, which allows tuning
    of the size of the backlog passed to the socket ``listen()``
    method.
    """

    proto = 'tcp'
    backlog = 1024

    def connect(self, target, acceptor, wrapper=None):
        """
        Initiate a connection from the tendril manager's endpoint.
        Once the connection is completed, a TCPTendril object will be
        created and passed to the given acceptor.

        :param target: The target of the connection attempt.
        :param acceptor: A callable which will initialize the state of
                         the new TCPTendril object.
        :param wrapper: A callable taking, as its first argument, a
                        socket.socket object.  The callable must
                        return a valid proxy for the socket.socket
                        object, which will subsequently be used to
                        communicate on the connection.

        For passing extra arguments to the acceptor or the wrapper,
        see the ``TendrilPartial`` class; for chaining together
        multiple wrappers, see the ``WrapperChain`` class.
        """

        # Call some common sanity-checks
        super(TCPTendrilManager, self).connect(target, acceptor, wrapper)

        # Set up the socket
        sock = socket.socket(self.addr_family, socket.SOCK_STREAM)

        with utils.SocketCloser(sock, ignore=[application.RejectConnection]):
            # Bind to our endpoint
            sock.bind(self.endpoint)

            # Connect to our target
            sock.connect(target)

            # Call any wrappers
            if wrapper:
                sock = wrapper(sock)

            # Now, construct a Tendril
            tend = TCPTendril(self, sock)

            # Finally, set up the application
            tend.application = acceptor(tend)

            # OK, let's track the tendril
            self._track_tendril(tend)

            # Start the tendril
            tend._start()

            # Might as well return the tendril, too
            return tend

        # The acceptor raised a RejectConnection exception, apparently
        sock.close()
        return None

    def listener(self, acceptor, wrapper):
        """
        Listens for new connections to the manager's endpoint.  Once a
        new connection is received, a TCPTendril object is generated
        for it and it is passed to the acceptor, which must initialize
        the state of the connection.  If no acceptor is given, no new
        connections can be initialized.

        :param acceptor: If given, specifies a callable that will be
                         called with each newly received TCPTendril;
                         that callable is responsible for initial
                         acceptance of the connection and for setting
                         up the initial state of the connection.  If
                         not given, no new connections will be
                         accepted by the TCPTendrilManager.
        :param wrapper: A callable taking, as its first argument, a
                        socket.socket object.  The callable must
                        return a valid proxy for the socket.socket
                        object, which will subsequently be used to
                        communicate on the connection.
        """

        # If we have no acceptor, there's nothing for us to do here
        if not acceptor:
            # Not listening on anything
            self.local_addr = None

            # Just sleep in a loop
            while True:
                gevent.sleep(600)
            return  # Pragma: nocover

        # OK, set up the socket
        sock = socket.socket(self.addr_family, socket.SOCK_STREAM)

        with utils.SocketCloser(sock):
            # Set up SO_REUSEADDR
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

            # Bind to our endpoint
            sock.bind(self.endpoint)

            # Get the assigned port number
            self.local_addr = sock.getsockname()

            # Call any wrappers
            if wrapper:
                sock = wrapper(sock)

            # Initiate listening
            sock.listen(self.backlog)

        # OK, now go into an accept loop with an error threshold of 10
        closer = utils.SocketCloser(sock, 10,
                                    ignore=[application.RejectConnection])
        while True:
            with closer:
                cli, addr = sock.accept()

                # OK, the connection has been accepted; construct a
                # Tendril for it
                tend = TCPTendril(self, cli, addr)

                # Set up the application
                with utils.SocketCloser(cli):
                    tend.application = acceptor(tend)

                    # Make sure we track the new tendril, but only if
                    # the acceptor doesn't throw any exceptions
                    self._track_tendril(tend)

                    # Start the tendril
                    tend._start()



## Copyright (C) 2012 by Kevin L. Mitchell <klmitch@mit.edu>
##
## This program is free software: you can redistribute it and/or
## modify it under the terms of the GNU General Public License as
## published by the Free Software Foundation, either version 3 of the
## License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program.  If not, see
## <http://www.gnu.org/licenses/>.

import abc
import struct


__all__ = ["FrameSwitch",
           "Framer",
           "IdentityFramer", "ChunkFramer", "LineFramer", "StructFramer",
           "StuffingFramer", "COBSFramer"]


class FrameSwitch(BaseException):
    """
    Sentinel exception used to inform the framer that it should hand
    back any remaining data.
    """

    pass


class FrameState(object):
    """
    Maintain state for framers.  The ``recv_buf`` and ``send_buf``
    attributes will always be present; framers will be given an
    opportunity to initialize other state attributes as necessary.
    Framers may use any attribute name except attributes beginning
    with '_'.
    """

    def __init__(self):
        """
        Initialize the FramerState object.  The ``recv_buf`` attribute
        will be initialized to the empty string, and the ``send_buf``
        attribute will be initialized to an empty list.
        """

        self.recv_buf = ''
        self.send_buf = []

        self._other = {}
        self._framer_id = None

    def __getattr__(self, name):
        """
        Retrieve a state attribute.
        """

        # Special treatment for internal attributes
        if name[0] == '_':
            raise AttributeError('%r object has no attribute %r' %
                                 (self.__class__.__name__, name))

        try:
            return self._other[name]
        except KeyError:
            raise AttributeError('%r object has no attribute %r' %
                                 (self.__class__.__name__, name))

    def __setattr__(self, name, value):
        """
        Set a state attribute.
        """

        # Special treatment for recv_buf and internal attributes
        if name in ('recv_buf', 'send_buf') or name[0] == '_':
            return super(FrameState, self).__setattr__(name, value)

        self._other[name] = value

    def __delattr__(self, name):
        """
        Delete a state attribute.
        """

        # Special treatment for recv_buf and internal attributes
        if name in ('recv_buf', 'send_buf') or name[0] == '_':
            raise AttributeError('%r object has no attribute %r' %
                                 (self.__class__.__name__, name))

        try:
            del self._other[name]
        except KeyError:
            raise AttributeError('%r object has no attribute %r' %
                                 (self.__class__.__name__, name))

    def _reset(self, framer):
        """
        Reset the state for the framer.  It is safe to call this
        method multiple times with the same framer; the ID of the
        framer object will be saved and the state only reset if the
        IDs are different.  After resetting the state, the framer's
        ``init_state()`` method will be called.
        """

        # Do nothing if we're already properly initialized
        if id(framer) == self._framer_id:
            return

        # Reset the state
        self._other = {}

        # Initialize the state and save the framer ID
        framer.init_state(self)
        self._framer_id = id(framer)


class Framer(object):
    """
    Abstract base class for all framers.  A framer is used by a stream
    transport, such as TCP, to chop the stream up into individual
    frames.  The ``buffer`` attribute will contain any partial frame
    data.
    """

    __metaclass__ = abc.ABCMeta

    def init_state(self, state):
        """Initialize the framer state."""

        pass

    @abc.abstractmethod
    def frameify(self, state, data):
        """
        Split data into individual frames.  Must buffer any
        unprocessed data in the ``state.recv_buf`` variable.  Each
        frame must be yielded as soon as it is carved out of the
        stream.

        The yield statement may throw a FrameSwitch exception, in
        which case any remaining unprocessed data must be immediately
        placed on the buffer and the generator exited.  This is used
        if the framer is changed.
        """

        yield  # Pragma: nocover

    @abc.abstractmethod
    def streamify(self, state, frame):
        """
        Convert a frame into stream data.  Returns the data.
        """

        pass  # Pragma: nocover


class IdentityFramer(Framer):
    """
    A framer for datagram transports, such as UDP.
    """

    def frameify(self, state, data):
        """Yield the data as a single frame."""

        try:
            yield state.recv_buf + data
        except FrameSwitch:
            pass
        finally:
            state.recv_buf = ''

    def streamify(self, state, frame):
        """Return the frame as data."""

        return frame


class ChunkFramer(IdentityFramer):
    """
    A framer for short-term streaming.  Once initialized with a given
    length, it yields received data as if it were a frame; however,
    once all the specified data has been accumulated, the remainder is
    pushed onto the receive buffer.
    """

    def __init__(self, chunk_len):
        """
        Initialize the ChunkFramer.

        :param chunk_len: The amount of data to pass through.
        """

        super(ChunkFramer, self).__init__()

        self.chunk_len = chunk_len

    def init_state(self, state):
        """Initialize the framer state."""

        state.chunk_remaining = self.chunk_len

    def frameify(self, state, data):
        """Yield chunk data as a single frame, and buffer the rest."""

        # If we've pulled in all the chunk data, buffer the data
        if state.chunk_remaining <= 0:
            state.recv_buf += data
            return

        # Pull in any partially-processed data
        data = state.recv_buf + data

        # Determine how much belongs to the chunk
        if len(data) <= state.chunk_remaining:
            chunk = data
            data = ''
        else:
            # Pull out only what's part of the chunk
            chunk = data[:state.chunk_remaining]
            data = data[state.chunk_remaining:]

        # Update the state
        state.recv_buf = data
        state.chunk_remaining -= len(chunk)

        # Yield the chunk
        try:
            yield chunk
        except FrameSwitch:
            pass


class LineFramer(Framer):
    """
    A line-oriented framer.  Frames are separated by newlines or by
    carriage return/newline pairs.  The line endings are stripped off.
    """

    def __init__(self, carriage_return=True):
        """
        Initialize the LineFramer.

        :param carriage_return: If ``True`` (the default), accept
                                carriage return/newline pairs as line
                                separators.  Also causes carriage
                                returns to be emitted.  If ``False``,
                                carriage returns are not stripped from
                                input and not emitted on output.
        """

        super(LineFramer, self).__init__()

        self.carriage_return = carriage_return
        self.line_end = '\r\n' if carriage_return else '\n'

    def frameify(self, state, data):
        """Split data into a sequence of lines."""

        # Pull in any partially-processed data
        data = state.recv_buf + data

        # Loop over the data
        while data:
            line, sep, rest = data.partition('\n')

            # Did we have a whole line?
            if sep != '\n':
                break

            # OK, update the data...
            data = rest

            # Now, strip off carriage return, if there is one
            if self.carriage_return and line[-1] == '\r':
                line = line[:-1]

            # Yield the line
            try:
                yield line
            except FrameSwitch:
                break

        # Put any remaining data back into the buffer
        state.recv_buf = data

    def streamify(self, state, frame):
        """Prepare frame for output as line-oriented data."""

        return '%s%s' % (frame, self.line_end)


class StructFramer(Framer):
    """
    A struct-oriented framer.  Uses the Python struct module to
    compose together a string length followed by the appropriate
    string.  Initialize with a string containing a single struct
    conversion string for integers.  The resulting packet will consist
    of the integer length of the frame, encoded in that format,
    followed by the frame itself.
    """

    def __init__(self, fmt):
        """
        Initialize the StructFramer.

        :param fmt: The struct-compliant format string for the integer
                    length of the frame.
        """

        # Sanity-check the fmt
        fmt_chr = None
        for c in fmt:
            if c in ('@', '=', '<', '>', '!', 'x'):
                # Modifiers and pads we can ignore
                continue

            if c in ('b', 'B', 'h', 'H', 'i', 'I', 'l', 'L', 'q', 'Q'):
                if fmt_chr:
                    raise ValueError("too many specifiers in format")
                fmt_chr = c
                continue

            # Invalid specifier
            raise ValueError("unrecognized specifier in format")

        if not fmt_chr:
            raise ValueError("no recognized specifier in format")

        super(StructFramer, self).__init__()

        self.fmt = struct.Struct(fmt)

    def init_state(self, state):
        """Initialize the framer state."""

        state.frame_len = None

    def frameify(self, state, data):
        """Split data into a sequence of frames."""

        # Pull in any partially-processed data
        data = state.recv_buf + data

        # Loop over the data
        while data:
            if state.frame_len is None:
                # Try to grab a frame length from the data
                if len(data) < self.fmt.size:
                    # Not enough data; try back later
                    break

                # Extract the length
                state.frame_len = self.fmt.unpack(data[:self.fmt.size])[0]
                data = data[self.fmt.size:]

            # Now that we have the frame length, extract the frame
            if len(data) < state.frame_len:
                # Not enough data; try back later
                break

            # OK, we have a full frame...
            frame = data[:state.frame_len]
            data = data[state.frame_len:]
            state.frame_len = None

            # Yield the frame
            try:
                yield frame
            except FrameSwitch:
                break

        # Put any remaining data back into the buffer
        state.recv_buf = data

    def streamify(self, state, frame):
        """Prepare frame for output as a length/frame stream."""

        return '%s%s' % (self.fmt.pack(len(frame)), frame)


class StuffingFramer(Framer):
    """
    A byte-stuffing framer.  Uses a technique of byte stuffing for
    separating frames.  In this technique, a sequence of bytes of a
    given length is used as a marker of the beginning of the frame,
    and another of the end of the frame.  Should the sequence appear
    inside the frame, an extra, throw-away byte is stuffed in to
    ensure the sequence does not match the begin-frame sequence.  One
    advantage of byte-stuffing is that it is easily possible to find
    the beginning of the next frame in the stream even if
    synchronization is momentarily lost.
    """

    def __init__(self, prefix='\xff' * 4, begin='\xff', end='\xfe', nop='\0'):
        """
        Initialize the StuffingFramer.

        :param prefix: A sequence of bytes which prefixes the
                       begin-frame and end-frame bytes.
        :param begin: A single byte which, when following the prefix
                      sequence, indicates the beginning of a frame.
        :param end: A single byte which, when following the prefix
                    sequence, indicates the end of a frame.
        :param nop: A single byte which, when following the prefix
                    sequence, is thrown away.  Used to interrupt
                    sequences internal to the frame which could be
                    interpreted as the beginning or ending of the
                    frame.
        """

        # Do a little sanity-checking
        if len(begin) != 1 or len(end) != 1 or len(nop) != 1:
            raise ValueError("begin, end, and nop must be length 1")
        elif nop == begin or nop == end:
            raise ValueError("nop must be distinct from begin and end")
        elif begin == end:
            raise ValueError("begin and end must be distinct")

        super(StuffingFramer, self).__init__()

        self.prefix = prefix
        self.begin = begin
        self.end = end
        self.nop = nop

    def init_state(self, state):
        """Initialize the framer state."""

        state.frame_start = False

    def frameify(self, state, data):
        """Split data into a sequence of frames."""

        # Pull in any partially-processed data
        data = state.recv_buf + data

        # Loop over the data
        while data:
            if not state.frame_start:
                try:
                    idx = data.index(self.prefix + self.begin)
                except ValueError:
                    # Can't find the start of a frame...
                    break

                # Advance data to the beginning of the frame (just
                # after the marker)
                data = data[idx + len(self.prefix) + 1:]
                state.frame_start = True

            # Now that data is sitting at the frame start, let's find
            # the frame ending
            try:
                idx = data.index(self.prefix + self.end)
            except ValueError:
                # Can't find the end of the frame...
                break

            # OK, extract the frame and advance the data
            frame = data[:idx]
            data = data[idx + len(self.prefix) + 1:]
            state.frame_start = (self.begin == self.end)

            # Now we need to unstuff the frame
            frame = self.prefix.join(frame.split(self.prefix + self.nop))

            # Yield the frame
            try:
                yield frame
            except FrameSwitch:
                break

        # Put any remaining data back into the buffer
        state.recv_buf = data

    def streamify(self, state, frame):
        """Prepare frame for output as a byte-stuffed stream."""

        # Split the frame apart for stuffing...
        pieces = frame.split(self.prefix)

        return '%s%s%s%s%s' % (self.prefix, self.begin,
                               (self.prefix + self.nop).join(pieces),
                               self.prefix, self.end)


class COBSFramer(Framer):
    """
    A byte-stuffing framer using the Consistent Overhead Byte Stuffing
    algorithm.  In COBS, data is encoded so as to eliminate the zero
    byte, making that byte available to indicate frame boundaries.
    The data itself is broken into blocks of data, each of which ends
    with a zero byte.  A variant that can more efficiently encode
    pairs of zero bytes can be activated by enabling zero pair
    elimination (the ``zpe`` parameter to the constructor).
    """

    _tabs = dict(enc_cobs={}, enc_cobs_zpe={},
                 dec_cobs={}, dec_cobs_zpe={})

    @classmethod
    def _get_tab(cls):
        """Generate and return the COBS table."""

        if not cls._tabs['dec_cobs']:
            # Compute the COBS table for decoding
            cls._tabs['dec_cobs']['\xff'] = (255, '')
            cls._tabs['dec_cobs'].update(dict((chr(l), (l, '\0'))
                                              for l in range(1, 255)))

            # Compute the COBS table for encoding
            cls._tabs['enc_cobs'] = [(255, '\xff'),
                                     dict((l, chr(l))
                                          for l in range(1, 255)),
                                     ]

        return cls._tabs['dec_cobs'], cls._tabs['enc_cobs']

    @classmethod
    def _get_tab_zpe(cls):
        """Generate and return the COBS ZPE table."""

        if not cls._tabs['dec_cobs_zpe']:
            # Compute the COBS ZPE table for decoding
            cls._tabs['dec_cobs_zpe']['\xe0'] = (224, '')
            cls._tabs['dec_cobs_zpe'].update(dict((chr(l), (l, '\0'))
                                                  for l in range(1, 224)))
            cls._tabs['dec_cobs_zpe'].update(dict((chr(l), (l - 224, '\0\0'))
                                                  for l in range(225, 256)))

            # Compute the COBS ZPE table for encoding
            cls._tabs['enc_cobs_zpe'] = [(224, '\xe0'),
                                         dict((l, chr(l))
                                              for l in range(1, 224)),
                                         dict((l - 224, chr(l))
                                              for l in range(225, 256))
                                         ]

        return cls._tabs['dec_cobs_zpe'], cls._tabs['enc_cobs_zpe']

    @staticmethod
    def _decode(frame, tab):
        """Decode a frame with the help of the table."""

        blocks = []

        # Decode each block
        while frame:
            length, endseq = tab[frame[0]]
            blocks.extend([frame[1:length], endseq])
            frame = frame[length:]

        # Remove one (and only one) trailing '\0' as necessary
        if blocks and len(blocks[-1]) > 0:
            blocks[-1] = blocks[-1][:-1]

        # Return the decoded plaintext
        return ''.join(blocks)

    def __init__(self, zpe=False):
        """
        Initialize the COBSFramer.

        :param zpe: If ``True``, enables the zero-pair elimination
                    variant of the COBS algorithm.
        """

        super(COBSFramer, self).__init__()

        self.zpe = zpe

    def frameify(self, state, data):
        """Split data into a sequence of frames."""

        # Pull in any partially-processed data
        data = state.recv_buf + data

        # Loop over the data
        while data:
            frame, sep, rest = data.partition('\0')

            # Did we have a whole frame?
            if sep != '\0':
                break

            # OK, update the data...
            data = rest

            # Now, decode the frame and yield it
            try:
                yield self._decode(frame, self._tables[0])
            except FrameSwitch:
                break

        # Put any remaining data back into the buffer
        state.recv_buf = data

    def streamify(self, state, frame):
        """Prepare frame for output as a COBS-encoded stream."""

        # Get the encoding table
        enc_tab = self._tables[1][:]

        # Need the special un-trailed block length and code
        untrail_len, untrail_code = enc_tab.pop(0)

        # Set up a repository to receive the encoded blocks
        result = []

        # Break the frame into blocks
        blocks = frame.split('\0')

        # Now, walk the block list; done carefully because we need
        # look-ahead in some cases
        skip = False
        for i in range(len(blocks)):
            # Skip handled blocks
            if skip:
                skip = False
                continue

            blk = blocks[i]

            # Encode un-trailed blocks
            while len(blk) >= untrail_len - 1:
                result.append(untrail_code + blk[:untrail_len - 1])
                blk = blk[untrail_len - 1:]

            # Do we care about look-ahead?
            if (len(enc_tab) > 1 and i + 1 < len(blocks) and
                    blocks[i + 1] == '' and len(blk) <= 30):
                # Use the second encoder table
                tab = enc_tab[1]

                # Skip the following empty block
                skip = True
            else:
                # Use the regular encoder table
                tab = enc_tab[0]

            # Encode the block
            result.append(tab[len(blk) + 1] + blk)

        # Stitch together the result blocks
        return ''.join(result) + '\0'

    @property
    def _tables(self):
        """Helper to retrieve the decoding and encoding tables."""

        return self._get_tab_zpe() if self.zpe else self._get_tab()



#!/usr/bin/env python

from setuptools import setup


def readreq(filename):
    result = []
    with open(filename) as f:
        for req in f:
            req = req.lstrip()
            if req.startswith('-e ') or req.startswith('http:'):
                idx = req.find('#egg=')
                if idx >= 0:
                    req = req[idx + 5:].partition('#')[0].strip()
                else:
                    pass
            else:
                req = req.partition('#')[0].strip()
            if not req:
                continue
            result.append(req)
    return result


def readfile(filename):
    with open(filename) as f:
        return f.read()


setup(
    name='Tendril',
    version='0.1.1',
    author='Kevin L. Mitchell',
    author_email='klmitch@mit.edu',
    url='http://github.com/klmitch/tendril',
    description='Frame-based Network Connection Tracker',
    long_description=readfile('README.rst'),
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Environment :: No Input/Output (Daemon)',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
        ],
    packages=['tendril'],
    install_requires=readreq('requirements.txt'),
    tests_require=readreq('test-requirements.txt'),
    entry_points={
        'tendril.manager': [
            'tcp = tendril.tcp:TCPTendrilManager',
            'udp = tendril.udp:UDPTendrilManager',
            ],
        },
    )



## Copyright (C) 2012 by Kevin L. Mitchell <klmitch@mit.edu>
##
## This program is free software: you can redistribute it and/or
## modify it under the terms of the GNU General Public License as
## published by the Free Software Foundation, either version 3 of the
## License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program.  If not, see
## <http://www.gnu.org/licenses/>.

import unittest

import gevent
from gevent import event
import mock
import pkg_resources

from tendril import manager


class ManagerForTest(manager.TendrilManager):
    proto = 'test'

    def connect(self, target, acceptor, wrapper=None):
        super(ManagerForTest, self).connect(target, acceptor, wrapper)

    def listener(self, acceptor, wrapper):
        pass


@mock.patch.dict(manager.TendrilManager._managers)
@mock.patch.dict(manager.TendrilManager._tendrils)
@mock.patch.dict(manager.TendrilManager._running_managers)
class TestTendrilManager(unittest.TestCase):
    def test_init(self):
        tm = ManagerForTest()

        self.assertEqual(tm.endpoint, ('', 0))
        self.assertEqual(tm.tendrils, {})
        self.assertEqual(tm.running, False)
        self.assertEqual(tm._local_addr, None)
        self.assertIsInstance(tm._local_addr_event, event.Event)
        self.assertFalse(tm._local_addr_event.is_set())
        self.assertEqual(tm._listen_thread, None)
        self.assertEqual(tm._manager_key, ('test', ('', 0)))
        self.assertEqual(dict(manager.TendrilManager._managers), {
            ('test', ('', 0)): tm,
        })
        self.assertEqual(manager.TendrilManager._tendrils, {})
        self.assertEqual(manager.TendrilManager._running_managers, {})

    def test_init_endpoint(self):
        tm = ManagerForTest(('127.0.0.1', 8080))

        self.assertEqual(tm.endpoint, ('127.0.0.1', 8080))
        self.assertEqual(tm.tendrils, {})
        self.assertEqual(tm.running, False)
        self.assertEqual(tm._local_addr, None)
        self.assertIsInstance(tm._local_addr_event, event.Event)
        self.assertFalse(tm._local_addr_event.is_set())
        self.assertEqual(tm._listen_thread, None)
        self.assertEqual(tm._manager_key, ('test', ('127.0.0.1', 8080)))
        self.assertEqual(dict(manager.TendrilManager._managers), {
            ('test', ('127.0.0.1', 8080)): tm,
        })
        self.assertEqual(manager.TendrilManager._tendrils, {})
        self.assertEqual(manager.TendrilManager._running_managers, {})

    def test_init_duplicate(self):
        dummy = mock.Mock()
        manager.TendrilManager._managers[('test', ('', 0))] = dummy

        self.assertRaises(ValueError, ManagerForTest)

    @mock.patch.object(pkg_resources, 'iter_entry_points')
    def test_get_manager_existing(self, mock_iter_entry_points):
        mock_manager = mock.Mock()
        manager.TendrilManager._managers[('test', ('', 0))] = mock_manager

        result = manager.get_manager('TEST')

        self.assertEqual(id(result), id(mock_manager))
        self.assertFalse(mock_iter_entry_points.called)

    @mock.patch.object(pkg_resources, 'iter_entry_points')
    def test_get_manager_noloader(self, mock_iter_entry_points):
        self.assertRaises(ValueError, manager.get_manager, 'test')
        mock_iter_entry_points.assert_called_once_with('tendril.manager',
                                                       'test')

    @mock.patch.object(pkg_resources, 'iter_entry_points')
    def test_get_manager_failedload(self, mock_iter_entry_points):
        mock_iter_entry_points.return_value = [
            mock.Mock(**{"load.side_effect": TypeError}),
        ]

        self.assertRaises(TypeError, manager.get_manager, 'test')
        mock_iter_entry_points.assert_called_once_with('tendril.manager',
                                                       'test')

    @mock.patch.object(pkg_resources, 'iter_entry_points')
    def test_get_manager(self, mock_iter_entry_points):
        loader = mock.Mock()
        mock_iter_entry_points.return_value = [
            mock.Mock(**{"load.side_effect": ImportError}),
            mock.Mock(**{"load.side_effect": pkg_resources.UnknownExtra}),
            mock.Mock(**{"load.return_value": loader}),
        ]

        manager.get_manager('test')

        mock_iter_entry_points.assert_called_once_with('tendril.manager',
                                                       'test')
        for m in mock_iter_entry_points.return_value:
            m.load.assert_called_once_with()
        loader.assert_called_once_with(('', 0))

    @mock.patch.object(pkg_resources, 'iter_entry_points')
    def test_get_manager_endpoint(self, mock_iter_entry_points):
        loader = mock.Mock()
        mock_iter_entry_points.return_value = [
            mock.Mock(**{"load.return_value": loader}),
        ]

        manager.get_manager('test', ('127.0.0.1', 8080))

        mock_iter_entry_points.assert_called_once_with('tendril.manager',
                                                       'test')
        for m in mock_iter_entry_points.return_value:
            m.load.assert_called_once_with()
        loader.assert_called_once_with(('127.0.0.1', 8080))

    def test_track_tendril(self):
        tm = ManagerForTest()
        tendril = mock.Mock(proto='test',
                            _tendril_key=(('127.0.0.1', 8080),
                                          ('127.0.0.2', 8880)))

        tm._track_tendril(tendril)

        self.assertEqual(tm.tendrils, {
            (('127.0.0.1', 8080), ('127.0.0.2', 8880)): tendril,
        })
        self.assertTrue('test' in manager.TendrilManager._tendrils)
        self.assertEqual(dict(manager.TendrilManager._tendrils['test']), {
            (('127.0.0.1', 8080), ('127.0.0.2', 8880)): tendril,
        })

    def test_untrack_tendril(self):
        tm = ManagerForTest()
        tendril = mock.Mock(proto='test',
                            _tendril_key=(('127.0.0.1', 8080),
                                          ('127.0.0.2', 8880)))
        tm.tendrils[tendril._tendril_key] = tendril
        manager.TendrilManager._tendrils['test'] = {
            tendril._tendril_key: tendril,
        }

        tm._untrack_tendril(tendril)

        self.assertEqual(tm.tendrils, {})
        self.assertTrue('test' in manager.TendrilManager._tendrils)
        self.assertEqual(manager.TendrilManager._tendrils['test'], {})

    def test_untrack_tendril_ignores_keyerror(self):
        tm = ManagerForTest()
        tendril = mock.Mock(proto='test',
                            _tendril_key=(('127.0.0.1', 8080),
                                          ('127.0.0.2', 8880)))

        tm._untrack_tendril(tendril)

        self.assertEqual(tm.tendrils, {})
        self.assertFalse('test' in manager.TendrilManager._tendrils)

    def test_contains(self):
        tm = ManagerForTest()
        tendril = mock.Mock(proto='test',
                            _tendril_key=(('127.0.0.1', 8080),
                                          ('127.0.0.2', 8880)))
        tm.tendrils[tendril._tendril_key] = tendril

        self.assertTrue((('127.0.0.1', 8080), ('127.0.0.2', 8880)) in tm)
        self.assertFalse((('127.0.0.1', 8080), ('127.0.0.3', 8880)) in tm)

    def test_getitem(self):
        tm = ManagerForTest()
        tendril = mock.Mock(proto='test',
                            _tendril_key=(('127.0.0.1', 8080),
                                          ('127.0.0.2', 8880)))
        tm.tendrils[tendril._tendril_key] = tendril

        result = tm[tendril._tendril_key]

        self.assertEqual(id(result), id(tendril))

    def test_find_tendril(self):
        tendril = mock.Mock(proto='test',
                            _tendril_key=(('127.0.0.1', 8080),
                                          ('127.0.0.2', 8880)))
        manager.TendrilManager._tendrils['test'] = {
            tendril._tendril_key: tendril,
        }

        result = manager.find_tendril('TEST', tendril._tendril_key)

        self.assertEqual(id(result), id(tendril))

    def test_connect_notrunning(self):
        tm = ManagerForTest()

        self.assertRaises(ValueError, tm.connect, ('127.0.0.1', 8080),
                          'acceptor')

    def test_connect_familymismatch(self):
        tm = ManagerForTest()
        tm.running = True

        self.assertRaises(ValueError, tm.connect, ('::1', 8080), 'acceptor')

    def test_connect_running(self):
        tm = ManagerForTest()
        tm.running = True

        tm.connect(('127.0.0.1', 8080), 'acceptor')

        # Note: verifying that an exception is not raised

    @mock.patch.object(gevent, 'spawn')
    def test_start_running(self, mock_spawn):
        tm = ManagerForTest()
        tm.running = True

        self.assertRaises(ValueError, tm.start)
        self.assertEqual(manager.TendrilManager._running_managers, {})
        self.assertFalse(mock_spawn.called)

    @mock.patch.object(gevent, 'spawn')
    def test_start_identical(self, mock_spawn):
        tm = ManagerForTest()
        manager.TendrilManager._running_managers[tm._manager_key] = tm

        self.assertRaises(ValueError, tm.start)
        self.assertEqual(manager.TendrilManager._running_managers, {
            tm._manager_key: tm,
        })
        self.assertFalse(mock_spawn.called)

    @mock.patch.object(gevent, 'spawn')
    def test_start(self, mock_spawn):
        tm = ManagerForTest()
        tm._local_addr = ('127.0.0.1', 8080)
        tm._local_addr_event.set()

        tm.start('acceptor', 'wrapper')

        self.assertEqual(tm.running, True)
        self.assertEqual(tm._local_addr, None)
        self.assertFalse(tm._local_addr_event.is_set())
        self.assertEqual(manager.TendrilManager._running_managers, {
            tm._manager_key: tm,
        })
        self.assertNotEqual(tm._listen_thread, None)
        mock_spawn.assert_called_once_with(tm.listener, 'acceptor', 'wrapper')
        tm._listen_thread.link.assert_called_once_with(tm.stop)

    def test_stop(self):
        tm = ManagerForTest()
        tm.running = True
        tm._local_addr = ('127.0.0.1', 8080)
        tm._local_addr_event.set()
        manager.TendrilManager._running_managers[tm._manager_key] = tm

        tm.stop()

        self.assertEqual(tm.running, False)
        self.assertEqual(tm._local_addr, None)
        self.assertFalse(tm._local_addr_event.is_set())
        self.assertEqual(manager.TendrilManager._running_managers, {})

    def test_stop_desync(self):
        tm = ManagerForTest()
        tm.running = True
        tm._local_addr = ('127.0.0.1', 8080)
        tm._local_addr_event.set()

        tm.stop()

        self.assertEqual(tm.running, False)
        self.assertEqual(tm._local_addr, None)
        self.assertFalse(tm._local_addr_event.is_set())
        self.assertEqual(manager.TendrilManager._running_managers, {})

    def test_shutdown(self):
        listen_thread = mock.Mock()
        tm = ManagerForTest()
        tm.running = True
        tm._local_addr = ('127.0.0.1', 8080)
        tm._local_addr_event.set()
        tm._listen_thread = listen_thread
        manager.TendrilManager._running_managers[tm._manager_key] = tm

        tm.shutdown()

        self.assertEqual(tm.tendrils, {})
        self.assertEqual(tm.running, False)
        self.assertEqual(tm._local_addr, None)
        self.assertFalse(tm._local_addr_event.is_set())
        self.assertEqual(tm._listen_thread, None)
        listen_thread.kill.assert_called_once_with()

    def test_shutdown_desync(self):
        listen_thread = mock.Mock()
        tm = ManagerForTest()
        tm.running = True
        tm._local_addr = ('127.0.0.1', 8080)
        tm._local_addr_event.set()
        tm._listen_thread = listen_thread

        tm.shutdown()

        self.assertEqual(tm.tendrils, {})
        self.assertEqual(tm.running, False)
        self.assertEqual(tm._local_addr, None)
        self.assertFalse(tm._local_addr_event.is_set())
        self.assertEqual(tm._listen_thread, None)
        listen_thread.kill.assert_called_once_with()

    def test_shutdown_close(self):
        listen_thread = mock.Mock()
        tendrils = [
            mock.Mock(proto='test',
                      _tendril_key=(('127.0.0.1', 8080),
                                    ('127.0.0.2', 8880))),
            mock.Mock(proto='test',
                      _tendril_key=(('127.0.0.1', 8080),
                                    ('127.0.0.3', 8880))),
            mock.Mock(proto='test',
                      _tendril_key=(('127.0.0.1', 8080),
                                    ('127.0.0.4', 8880))),
        ]

        tm = ManagerForTest()
        tm.running = True
        tm._local_addr = ('127.0.0.1', 8080)
        tm._local_addr_event.set()
        tm.tendrils.update(dict((t._tendril_key, t) for t in tendrils))
        tm._listen_thread = listen_thread
        manager.TendrilManager._running_managers[tm._manager_key] = tm

        tm.shutdown()

        self.assertEqual(tm.tendrils, {})
        self.assertEqual(tm.running, False)
        self.assertEqual(tm._local_addr, None)
        self.assertFalse(tm._local_addr_event.is_set())
        self.assertEqual(tm._listen_thread, None)
        listen_thread.kill.assert_called_once_with()
        for tend in tendrils:
            tend.close.assert_called_once_with()

    def test_get_local_addr_notrunning(self):
        tm = ManagerForTest()
        tm._local_addr = ('127.0.0.1', 8080)
        tm._local_addr_event = mock.Mock()

        result = tm.get_local_addr()

        self.assertEqual(result, None)
        self.assertFalse(tm._local_addr_event.wait.called)

    def test_get_local_addr_notset(self):
        tm = ManagerForTest()
        tm.running = True
        tm._local_addr = ('127.0.0.1', 8080)
        tm._local_addr_event = mock.Mock(**{
            'wait.return_value': False,
        })

        result = tm.get_local_addr('timeout')

        self.assertEqual(result, None)
        tm._local_addr_event.wait.assert_called_once_with('timeout')

    def test_get_local_addr(self):
        tm = ManagerForTest()
        tm.running = True
        tm._local_addr = ('127.0.0.1', 8080)
        tm._local_addr_event = mock.Mock(**{
            'wait.return_value': True,
        })

        result = tm.get_local_addr()

        self.assertEqual(result, ('127.0.0.1', 8080))
        tm._local_addr_event.wait.assert_called_once_with(None)

    @mock.patch.object(manager.TendrilManager, 'get_local_addr',
                       return_value=('127.0.0.1', 8080))
    def test_local_addr_getter(self, mock_get_local_addr):
        tm = ManagerForTest()

        self.assertEqual(tm.local_addr, ('127.0.0.1', 8080))

    def test_local_addr_setter(self):
        tm = ManagerForTest()
        tm._local_addr_event = mock.Mock()

        tm.local_addr = ('127.0.0.1', 8080)

        self.assertEqual(tm._local_addr, ('127.0.0.1', 8080))
        tm._local_addr_event.set.assert_called_once_with()



## Copyright (C) 2012 by Kevin L. Mitchell <klmitch@mit.edu>
##
## This program is free software: you can redistribute it and/or
## modify it under the terms of the GNU General Public License as
## published by the Free Software Foundation, either version 3 of the
## License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program.  If not, see
## <http://www.gnu.org/licenses/>.

import abc
import weakref

import gevent
from gevent import event
import pkg_resources

from tendril import utils


__all__ = ["get_manager", "find_tendril"]


class TendrilManager(object):
    """
    Manages all connections through a particular endpoint.  Handles
    accepting new connections, creating new outgoing connections, and
    buffering data to and from the network in the case of network
    protocols that do not provide connection state (i.e., UDP).  This
    is an abstract base class; see the tcp.py and udp.py files for
    implementations.

    The TendrilManager wraps connection state in a Tendril object,
    which provides the necessary information to maintain state for the
    connection--including application-provided state.  Application
    state should subclass the tendril.ApplicationState class.
    """

    __metaclass__ = abc.ABCMeta

    _managers = weakref.WeakValueDictionary()
    _tendrils = {}
    _running_managers = {}

    @classmethod
    def get_manager(cls, proto, endpoint=None):
        """
        Find a manager matching the given protocol and endpoint.  If
        no matching manager currently exists, creates a new one.
        (Manager classes are looked up using the ``tendril.manager``
        entrypoint, and the name of the entrypoint corresponds to the
        ``proto``.  This method makes no guarantees about whether the
        manager is running; make sure to check the ``running``
        attribute and call the ``start()`` method if necessary.

        :param proto: The underlying network protocol, such as "tcp"
                      or "udp".
        :param endpoint: Identifies the endpoint of the
                         TendrilManager.  This will be the IP address
                         and port number, as a tuple, on which to
                         accept connections and from which to initiate
                         connections.  If not given, defaults to
                         ``('', 0)``.

        Note: TendrilManager instances can only support one of IPv4 or
        IPv6, and outgoing connection addresses must match how the
        TendrilManager was initialized.  To get a TendrilManager
        supporting outgoing IPv6 connections, use the endpoint
        ``('::', 0)``.
        """

        # First, normalize the proto and endpoint
        proto = proto.lower()
        endpoint = endpoint or ('', 0)

        # See if the manager already exists
        if (proto, endpoint) in cls._managers:
            return cls._managers[(proto, endpoint)]

        # OK, need to create a new one; use pkg_resources
        for ep in pkg_resources.iter_entry_points('tendril.manager', proto):
            try:
                manager_cls = ep.load()
                break
            except (ImportError, pkg_resources.UnknownExtra):
                continue
        else:
            raise ValueError("unknown protocol %r" % proto)

        return manager_cls(endpoint)

    @classmethod
    def find_tendril(cls, proto, addr):
        """
        Finds the tendril corresponding to the protocol and address
        tuple.  Returns the Tendril object, or raises KeyError if the
        tendril is not tracked.

        The address tuple is the tuple of the local address and the
        remote address for the tendril.
        """

        # First, normalize the proto
        proto = proto.lower()

        # Now, find and return the tendril
        return cls._tendrils[proto][addr]

    def __init__(self, endpoint=None):
        """
        Initialize a TendrilManager.

        :param endpoint: Identifies the endpoint of the
                         TendrilManager.  This will be the IP address
                         and port number, as a tuple, on which to
                         accept connections and from which to initiate
                         connections.  If not given, defaults to
                         ``('', 0)``.

        Note: TendrilManager instances can only support one of IPv4 or
        IPv6, and outgoing connection addresses must match how the
        TendrilManager was initialized.  To get a TendrilManager
        supporting outgoing IPv6 connections, use the endpoint
        ``('::', 0)``.
        """

        self.endpoint = endpoint or ('', 0)
        self.addr_family = utils.addr_info(self.endpoint)
        self.tendrils = {}
        self.running = False
        self._local_addr = None
        self._local_addr_event = event.Event()
        self._local_addr_event.clear()

        self._listen_thread = None

        # Make sure we don't already exist...
        if self._manager_key in self._managers:
            raise ValueError("Identical TendrilManager already exists")

        # Save a reference to ourself
        self._managers[self._manager_key] = self

    def __contains__(self, addr):
        """
        Finds if a tendril corresponding to the address tuple exists.
        Returns ``True`` if it does or ``False`` otherwise.

        The address tuple is the tuple of the local address and the
        remote address for the tendril.
        """

        return addr in self.tendrils

    def __getitem__(self, addr):
        """
        Finds the tendril corresponding to the address tuple.  Returns
        the Tendril object, or raises KeyError if the tendril is not
        tracked by this TendrilManager.

        The address tuple is the tuple of the local address and the
        remote address for the tendril.
        """

        return self.tendrils[addr]

    def _track_tendril(self, tendril):
        """
        Adds the tendril to the set of tracked tendrils.
        """

        self.tendrils[tendril._tendril_key] = tendril

        # Also add to _tendrils
        self._tendrils.setdefault(tendril.proto, weakref.WeakValueDictionary())
        self._tendrils[tendril.proto][tendril._tendril_key] = tendril

    def _untrack_tendril(self, tendril):
        """
        Removes the tendril from the set of tracked tendrils.
        """

        try:
            del self.tendrils[tendril._tendril_key]
        except KeyError:
            pass

        # Also remove from _tendrils
        try:
            del self._tendrils[tendril.proto][tendril._tendril_key]
        except KeyError:
            pass

    def start(self, acceptor=None, wrapper=None):
        """
        Starts the TendrilManager.

        :param acceptor: If given, specifies a callable that will be
                         called with each newly received Tendril;
                         that callable is responsible for initial
                         acceptance of the connection and for setting
                         up the initial state of the connection.  If
                         not given, no new connections will be
                         accepted by the TendrilManager.
        :param wrapper: A callable taking, as its first argument, a
                        socket.socket object.  The callable must
                        return a valid proxy for the socket.socket
                        object, which will subsequently be used to
                        communicate on the connection.

        For passing extra arguments to the acceptor or the wrapper,
        see the ``TendrilPartial`` class; for chaining together
        multiple wrappers, see the ``WrapperChain`` class.
        """

        # Don't allow a double-start
        if self.running:
            raise ValueError("TendrilManager already running")

        # Look out for conflicts
        if self._manager_key in self._running_managers:
            raise ValueError("Identical TendrilManager already exists")

        # In a moment, we will begin running
        self.running = True
        self._local_addr = None
        self._local_addr_event.clear()

        # Add ourself to the dictionary of running managers
        self._running_managers[self._manager_key] = self

        # Start the listening thread
        self._listen_thread = gevent.spawn(self.listener, acceptor,
                                           wrapper)

        # Make sure to reset running if it exits
        self._listen_thread.link(self.stop)

    def stop(self, *args):
        """
        Stops the TendrilManager.  Requires cooperation from the
        listener implementation, which must watch the ``running``
        attribute and ensure that it stops accepting connections
        should that attribute become False.  Note that some tendril
        managers will not exit from the listening thread until all
        connections have been closed.
        """

        # Remove ourself from the dictionary of running managers
        try:
            del self._running_managers[self._manager_key]
        except KeyError:
            pass

        self.running = False
        self._local_addr = None
        self._local_addr_event.clear()

    def shutdown(self):
        """
        Unconditionally shuts the TendrilManager down, killing all
        threads and closing all tendrils.
        """

        # Remove ourself from the dictionary of running managers
        try:
            del self._running_managers[self._manager_key]
        except KeyError:
            pass

        # Kill the listening thread
        self._listen_thread.kill()

        # Close all the connections
        for conn in self.tendrils.values():
            conn.close()

        # Ensure all data is appropriately reset
        self.tendrils = {}
        self.running = False
        self._local_addr = None
        self._local_addr_event.clear()
        self._listen_thread = None

    def get_local_addr(self, timeout=None):
        """
        Retrieve the current local address.

        :param timeout: If not given or given as ``None``, waits until
                        the local address is available.  Otherwise,
                        waits for as long as specified.  If the local
                        address is not set by the time the timeout
                        expires, returns ``None``.
        """

        # If we're not running, just return None
        if not self.running:
            return None

        # OK, we're running; wait on the _local_addr_event
        if not self._local_addr_event.wait(timeout):
            # Still not set after timeout
            return None

        # We have a local address!
        return self._local_addr

    @property
    def _manager_key(self):
        """
        Returns a unique key identifying this manager.
        """

        return (self.proto, self.endpoint)

    @property
    def local_addr(self):
        """
        Retrieve the local address the manager is listening on.
        """

        return self.get_local_addr()

    @local_addr.setter
    def local_addr(self, value):
        """
        Set the local address the manager is listening on.  Notifies
        all waiters that the address is available.
        """

        self._local_addr = value
        self._local_addr_event.set()

    @abc.abstractmethod
    def connect(self, target, acceptor, wrapper=None):
        """
        Initiate a connection from the tendril manager's endpoint.
        Once the connection is completed, a Tendril object will be
        created and passed to the given acceptor.

        :param target: The target of the connection attempt.
        :param acceptor: A callable which will initialize the state of
                         the new Tendril object.
        :param wrapper: A callable taking, as its first argument, a
                        socket.socket object.  The callable must
                        return a valid proxy for the socket.socket
                        object, which will subsequently be used to
                        communicate on the connection.

        For passing extra arguments to the acceptor or the wrapper,
        see the ``TendrilPartial`` class; for chaining together
        multiple wrappers, see the ``WrapperChain`` class.
        """

        if not self.running:
            raise ValueError("TendrilManager not running")

        # Check the target address
        fam = utils.addr_info(target)

        # Verify that we're in the right family
        if self.addr_family != fam:
            raise ValueError("address family mismatch")

    @abc.abstractmethod
    def listener(self, acceptor, wrapper):
        """
        Listens for new connections to the manager's endpoint.  Once a
        new connection is received, a Tendril object is generated
        for it and it is passed to the acceptor, which must initialize
        the state of the connection.  If no acceptor is given, no new
        connections can be initialized, but some tendril managers
        still need a listening thread anyway.

        :param acceptor: If given, specifies a callable that will be
                         called with each newly received Tendril;
                         that callable is responsible for initial
                         acceptance of the connection and for setting
                         up the initial state of the connection.  If
                         not given, no new connections will be
                         accepted by the TendrilManager.
        :param wrapper: A callable taking, as its first argument, a
                        socket.socket object.  The callable must
                        return a valid proxy for the socket.socket
                        object, which will subsequently be used to
                        communicate on the connection.
        """

        pass  # Pragma: nocover

    @abc.abstractproperty
    def proto(self):
        """Retrieve the name of the underlying network protocol."""

        pass  # Pragma: nocover


get_manager = TendrilManager.get_manager
find_tendril = TendrilManager.find_tendril
